/* $Id: cpl_image_filter_body.h,v 1.68 2010/11/11 09:23:18 llundin Exp $
 *
 * This file is part of the ESO Common Pipeline Library
 * Copyright (C) 2001-2008 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

static void
ADDTYPE_TWO(cpl_filter_average)(void *, const void *, int, int, int,
                                int, cpl_border_mode);


static void
ADDTYPE_TWO(cpl_filter_average_bpm)(void *, cpl_binary **,
                                    const void *, const cpl_binary *,
                                    int, int, int, int, cpl_border_mode);

static void
ADDTYPE_TWO(cpl_filter_average_slow)(void *, cpl_binary **,
                                     const void *, const cpl_binary *,
                                     const cpl_binary *, int, int, int, int,
                                     cpl_border_mode);

static void
ADDTYPE_TWO(cpl_filter_stdev)(void *, const void *, int, int, int, int,
                              cpl_border_mode);

static void
ADDTYPE_TWO(cpl_filter_stdev_slow)(void *, cpl_binary **,
                                   const void *, const cpl_binary *,
                                   const cpl_binary *, int, int, int, int,
                                   cpl_border_mode);

static void
ADDTYPE_TWO(cpl_filter_linear_slow)(void *, cpl_binary **,
                                    const void *, const cpl_binary *,
                                    const double *, int, int, int, int,
                                    cpl_border_mode);

static void
ADDTYPE_TWO(cpl_filter_morpho_slow)(void *, cpl_binary **,
                                    const void *, const cpl_binary *,
                                    const double *, int, int, int, int,
                                    cpl_border_mode);

#ifndef CPL_FILTER_NO_RECIPROCAL
#ifndef ACCU_TYPE_IS_INT
#define ALLOW_RECIPROCAL
#endif
#endif

#ifdef ALLOW_RECIPROCAL
  /* If possible, and in order to speed-up the code, the central part of
     averaged image is done with multiplication instead of division, incurring
     one extra round-off */
#define RUNSUM_MEAN (runsum * rnpix)
#else
#define RUNSUM_MEAN (runsum / npix)
#endif


/*
  The functionss below are documented with Doxygen as usual,
  but the preprocessor generated function names mean that
  doxygen cannot actually parse it. So it is all ignored.
  @cond
 */

#ifdef IN_EQ_OUT

/* Start of no-cast code */

static void ADDTYPE(cpl_filter_median_slow)(void *, cpl_binary **,
                                            const void *,
                                            const cpl_binary *,
                                            const cpl_binary *,
                                            int, int, int,
                                            int, cpl_border_mode);

/*----------------------------------------------------------------------------*/
/**
   @internal
   @ingroup cpl_image
   @brief  Median filter with bad pixels
   @param  vout   The pixel buffer to hold the filtered result
   @param  ppbpm *ppbpm is the bpm, or NULL if still to be created
   @param  vin    The input pixel buffer to be filtered
   @param  bpm   The input bad pixel map, or NULL if none
   @param  pmask The window
   @param  nx    image width
   @param  ny    image height
   @param  rx    filtering half-size in x
   @param  ry    filtering half-size in y
   @return void
   
*/
/*----------------------------------------------------------------------------*/
static void
ADDTYPE(cpl_filter_median_slow)(void * vout, cpl_binary ** ppbpm,
                                const void * vin,
                                const cpl_binary * bpm,
                                const cpl_binary * pmask,
                                int nx, int ny,
                                int rx, int ry,
                                cpl_border_mode mode)
{

    if (mode == CPL_BORDER_FILTER) {
        OUT_TYPE * out = (OUT_TYPE*)vout;
        const IN_TYPE  * in  = (const IN_TYPE*)vin;
        IN_TYPE window[(2 * rx + 1) * (2 * ry + 1)];
        int     x, y;

        for (y = 0; y < ny; y++) {
            const int y_1 = CX_MAX(0,    y - ry);
            const int y2  = CX_MIN(ny-1, y + ry);
            for (x = 0; x < nx; x++) {
                const int          x1 = CX_MAX(0,    x - rx);
                const int          x2 = CX_MIN(nx-1, x + rx);
                const IN_TYPE    * inj  = in + y_1 * nx;
                const cpl_binary * bpmj = bpm ? bpm + y_1 * nx : NULL;
                const cpl_binary * pmaskj = pmask
                     + (rx - x) + ((ry - y) + y_1) * (2 * rx + 1);
                int k = 0;
                int i, j;

                for (j = y_1; j <= y2; j++, inj += nx, bpmj += nx,
                         pmaskj += 2 * rx + 1) {
                    for (i = x1; i <= x2; i++) {
                        if ((bpm == NULL || !bpmj[i]) && pmaskj[i])
                            window[k++] = inj[i];
                    }
                }
                if (k > 0) {
                    out[x + y * nx] = ADDTYPE(cpl_tools_get_median)(window, k);
                } else {
                    /* Flag as bad - and set to zero */
                    if (*ppbpm == NULL) {
                        *ppbpm = cpl_calloc(nx * ny, sizeof(cpl_binary));
                    }
                    (*ppbpm)[x + y * nx] = CPL_BINARY_1;
                    out[x + y * nx] = (OUT_TYPE)0;
                }
            }
        }
    } else {
        /* FIXME: Support more modes ? */
        (void)cpl_error_set(cpl_func, CPL_ERROR_UNSUPPORTED_MODE);
    }

    return;
}

/* End of no-cast code */
#endif

/*----------------------------------------------------------------------------*/
/**
   @internal
   @ingroup cpl_image
   @brief  Average filter with bad pixels
   @param  vout  The pixel buffer to hold the filtered result
   @param  ppbpm *ppbpm is the bpm, or NULL if still to be created
   @param  vin   The input pixel buffer to be filtered
   @param  bpm   The input bad pixel map, or NULL if none
   @param  pmask The window
   @param  nx    image width
   @param  ny    image height
   @param  rx    filtering half-size in x
   @param  ry    filtering half-size in y
   @return void
   @see cpl_filter_median_slow
   @note FIXME: Modified from cpl_filter_median_slow
   
*/
/*----------------------------------------------------------------------------*/
static void
ADDTYPE_TWO(cpl_filter_average_slow)(void * vout, cpl_binary ** ppbpm,
                                     const void * vin,
                                     const cpl_binary * bpm,
                                     const cpl_binary * pmask,
                                     int nx, int ny,
                                     int rx, int ry,
                                     cpl_border_mode mode)
{

    if (mode == CPL_BORDER_FILTER) {
        OUT_TYPE * out = (OUT_TYPE*)vout;
        const IN_TYPE  * in  = (const IN_TYPE*)vin;
        int        x, y;

        for (y = 0; y < ny; y++) {
            const int y_1 = CX_MAX(0,    y - ry);
            const int y2  = CX_MIN(ny-1, y + ry);
            for (x = 0; x < nx; x++) {
                const int          x1 = CX_MAX(0,    x - rx);
                const int          x2 = CX_MIN(nx-1, x + rx);
                const IN_TYPE   * inj  = in + y_1 * nx;
                const cpl_binary * bpmj = bpm ? bpm + y_1 * nx : NULL;
                const cpl_binary * pmaskj = pmask
                     + (rx - x) + ((ry - y) + y_1) * (2 * rx + 1);
                ACCU_TYPE meansum = (ACCU_TYPE)0; /* Sum for int else mean */

                int k = 0;
                int i, j;

                for (j = y_1; j <= y2; j++, inj += nx, bpmj += nx,
                         pmaskj += 2 * rx + 1) {
                    for (i = x1; i <= x2; i++) {
                        if ((bpm == NULL || !bpmj[i]) && pmaskj[i]) {
#if defined ACCU_TYPE_IS_INT || !defined CPL_FILTER_AVERAGE_SLOW
                            meansum += (ACCU_TYPE)inj[i];
                            k++;
#else
                            /* This recurrence is taken from
                               cpl_vector_get_mean() */
                            /* It is turned off by default, since
                               the number of samples is limited */
                            meansum += (inj[i] - meansum) / ++k;
#endif
                        }
                    }
                }
                if (k > 0) {
#if defined ACCU_TYPE_IS_INT || !defined CPL_FILTER_AVERAGE_SLOW
                    meansum /= k;
#endif
                } else {
                    /* Flag as bad - and set to zero */
                    if (*ppbpm == NULL) {
                        *ppbpm = cpl_calloc(nx * ny, sizeof(cpl_binary));
                    }
                    (*ppbpm)[x + y * nx] = CPL_BINARY_1;
                }
                out[x + y * nx] = (OUT_TYPE)meansum;
            }
        }
    } else {
        /* FIXME: Support more modes ? */
        (void)cpl_error_set(cpl_func, CPL_ERROR_UNSUPPORTED_MODE);
    }

    return;
}

/*----------------------------------------------------------------------------*/
/**
   @internal
   @ingroup cpl_image
   @brief  Stdev filter with bad pixels
   @param  vout  The pixel buffer to hold the filtered result
   @param  ppbpm *ppbpm is the bpm, or NULL if still to be created
   @param  vin   The input pixel buffer to be filtered
   @param  bpm   The input bad pixel map, or NULL if none
   @param  pmask The window
   @param  nx    image width
   @param  ny    image height
   @param  rx    filtering half-size in x
   @param  ry    filtering half-size in y
   @return void
   @see cpl_filter_average_slow, cpl_tools_get_variancesum
   @note FIXME: Modified from cpl_filter_average_slow
   
*/
/*----------------------------------------------------------------------------*/
static void
ADDTYPE_TWO(cpl_filter_stdev_slow)(void * vout, cpl_binary ** ppbpm,
                                    const void * vin,
                                    const cpl_binary * bpm,
                                    const cpl_binary * pmask,
                                    int nx, int ny,
                                    int rx, int ry,
                                    cpl_border_mode mode)
{

    if (mode == CPL_BORDER_FILTER) {
        OUT_TYPE * out = (OUT_TYPE*)vout;
        const IN_TYPE  * in  = (const IN_TYPE*)vin;
        int        x, y;

        for (y = 0; y < ny; y++) {
            const int y_1 = CX_MAX(0,    y - ry);
            const int y2  = CX_MIN(ny-1, y + ry);
            for (x = 0; x < nx; x++) {
                const int          x1 = CX_MAX(0,    x - rx);
                const int          x2 = CX_MIN(nx-1, x + rx);
                const IN_TYPE   * inj  = in + y_1 * nx;
                const cpl_binary * bpmj = bpm ? bpm + y_1 * nx : NULL;
                const cpl_binary * pmaskj = pmask
                     + (rx - x) + ((ry - y) + y_1) * (2 * rx + 1);
                ACCU_FLOATTYPE varsum = 0.0;
                ACCU_FLOATTYPE mean = 0.0;

                double k = 0;
                int i, j;

                for (j = y_1; j <= y2; j++, inj += nx, bpmj += nx,
                         pmaskj += 2 * rx + 1) {
                    for (i = x1; i <= x2; i++) {
                        if ((bpm == NULL || !bpmj[i]) && pmaskj[i]) {
                            const ACCU_FLOATTYPE delta
                                = (ACCU_FLOATTYPE)inj[i] - mean;

                            varsum += k * delta * delta / (k + 1.0);
                            mean   += delta / (k + 1.0);
                            k += 1.0;
                        }
                    }
                }
                if (k > 1) {
                    out[x + y * nx] = (OUT_TYPE)sqrt(varsum / (k - 1.0));
                } else {
                    /* Flag as bad - and set to zero */
                    if (*ppbpm == NULL) {
                        *ppbpm = cpl_calloc(nx * ny, sizeof(cpl_binary));
                    }
                    (*ppbpm)[x + y * nx] = CPL_BINARY_1;
                    out[x + y * nx] = (OUT_TYPE)0;
                }
            }
        }
    } else {
        /* FIXME: Support more modes ? */
        (void)cpl_error_set(cpl_func, CPL_ERROR_UNSUPPORTED_MODE);
    }

    return;
}

/*----------------------------------------------------------------------------*/
/**
   @internal
   @ingroup cpl_image
   @brief  Stdev filter without bad pixels
   @param  vout  The pixel buffer to hold the filtered result
   @param  vin   The input pixel buffer to be filtered
   @param  nx    image width
   @param  ny    image height
   @param  rx    filtering half-size in x
   @param  ry    filtering half-size in y
   @return void
   @see cpl_filter_stdev_slow
   
*/
/*----------------------------------------------------------------------------*/
static void
ADDTYPE_TWO(cpl_filter_stdev)(void * vout,
                              const void * vin,
                              int nx, int ny,
                              int hsizex, int hsizey,
                              cpl_border_mode border_mode)
{

    OUT_TYPE * out = (OUT_TYPE*)vout;
    const IN_TYPE  * in  = (const IN_TYPE*)vin;

    double * colvarsum;
    double * colmean;
    int npix, npixy;
    int x, y, i;

    double npixcol, ncol;

    assert(out != NULL);
    assert(in != NULL);

    assert( hsizex >= 0 );
    assert( hsizey >= 0 );
    assert( 2*hsizex + 1 <= nx ); /* Caller does not (need to) support this */
    assert( 2*hsizey + 1 <= ny ); /* Caller does not (need to) support this */

    assert( hsizex > 0 || hsizey > 0);

    /* FIXME: Support more modes ? */
    assert( border_mode == CPL_BORDER_FILTER );

    /* The number of pixels prior to the first stdev */
    npixy = hsizey;
    npix = (1+hsizex) * npixy;

    /* Initialize the column sums */
    colvarsum = cpl_calloc(nx, sizeof(double));
    colmean   = cpl_calloc(nx, sizeof(double));

    npixcol = 0.0;

    for (y=0; y < hsizey; y++) {
        for (i=0; i < nx; i++) {
        const double delta = (double)in[i + y * nx] - colmean[i];

        colvarsum[i] += npixcol / (npixcol + 1.0) * delta * delta;
        colmean[i]   += delta / (npixcol + 1.0);
        }

    npixcol += 1.0;
    }

    
    /* Loop over all the first rows, that cannot be covered by the kernel */
    for (y = 0; y <= hsizey; y++) {

        double runvarsum = (double)0;
        double runmean   = (double)0;

#ifdef ALLOW_RECIPROCAL
        double rnpix;
#endif

        npixy ++;
        npix = (hsizex)*npixy;
    
    ncol = 0.0;

        for (x = 0; x < 1; x ++) {

            /* Update the first 1 + hsizex column sums */
            for (i = 0; i <= hsizex; i ++) {
        const double delta    = 
            (double)in[i + (y + hsizey) * nx] - colmean[i];
        double rundelta;
 
        colvarsum[i] += npixcol / (npixcol + 1.0) * delta * delta;
        colmean[i]   += delta / (npixcol + 1.0);

        rundelta = colmean[i] - runmean;

        runvarsum += 
            colvarsum[i] + ncol / (ncol + 1.0) * (npixcol + 1.0) *
            rundelta * rundelta;
        runmean   += rundelta / (ncol + 1.0);
        
        ncol += 1.0;
        }

            npix += npixy;

            out[x + y * nx] = (OUT_TYPE)sqrt(runvarsum/(npix - 1.0));
        }

        for (; x <= hsizex; x ++) {
        const double delta = 
        (double)in[x + hsizex + (y + hsizey) * nx] - 
        colmean[x + hsizex];
        double rundelta;
        
        colvarsum[x + hsizex] += npixcol / (npixcol + 1.0) * delta * delta;
        colmean[x + hsizex]   += delta / (npixcol + 1.0);

        rundelta = colmean[x + hsizex] - runmean;

        runvarsum += 
        colvarsum[x + hsizex] + ncol / (ncol + 1.0) * (npixcol + 1.0) *
        rundelta * rundelta;
        runmean   += rundelta / (ncol + 1.0);

        ncol += 1.0;
            npix += npixy;

            out[x + y * nx] = (OUT_TYPE)sqrt(runvarsum/(npix - 1.0));
        }

#ifdef ALLOW_RECIPROCAL
        rnpix = (double)1.0/npix;
#endif

        for (; x < nx-hsizex; x ++) {
        const double delta = 
        (double)in[x + hsizex + (y + hsizey) * nx] - 
        colmean[x + hsizex];
        double rundelta;
        
        colvarsum[x + hsizex] += npixcol / (npixcol + 1.0) * delta * delta;
        colmean[x + hsizex]   += delta / (npixcol + 1.0);

        if (ncol > 1.0) {
        rundelta = colmean[x - hsizex - 1] - runmean;

        runvarsum -= 
            colvarsum[x - hsizex - 1] + 
            ncol / (ncol - 1.0) * (npixcol + 1.0) *
            rundelta * rundelta;
        runmean   -= rundelta / (ncol - 1.0);

        rundelta = colmean[x + hsizex] - runmean;

        runvarsum += 
            colvarsum[x + hsizex] + 
            (ncol - 1.0) / ncol * (npixcol + 1.0) *
            rundelta * rundelta;
        runmean   += rundelta / ncol;
        } else {
        runvarsum = colvarsum[x + hsizex];
        }

            out[x + y * nx] = (OUT_TYPE)sqrt(runvarsum/(npix - 1.0));
        }

        for (; x < nx; x++) {
        const double rundelta = colmean[x - hsizex - 1] - runmean;

        runvarsum -= 
        colvarsum[x - hsizex - 1] + 
        ncol / (ncol - 1.0) * (npixcol + 1.0) *
        rundelta * rundelta;
        runmean   -= rundelta / (ncol - 1.0);

        ncol -= 1.0;
            npix -= npixy;

            out[x + y * nx] = (OUT_TYPE)sqrt(runvarsum/(npix - 1.0));
        }

    npixcol += 1.0;
    }

    /* Loop over the central rows */
    for (;y < ny-hsizey; y++) {

        double runvarsum = (double)0;
        double runmean   = (double)0;

#ifdef ALLOW_RECIPROCAL
        double rnpix;
#endif
        npix = (hsizex)*npixy;

    ncol = 0.0;

        for (x = 0; x < 1; x++) {

            /* Update the first 1 + hsizex column sums */
            for (i = 0; i <= hsizex; i ++) {
        double delta;
        double rundelta;

        if (npixcol > 1.0) {
            delta = (double)in[i + (y - hsizey - 1) * nx] - colmean[i];
 
            colvarsum[i] -= 
            npixcol / (npixcol - 1.0) * delta * delta;
            colmean[i]   -= delta / (npixcol - 1.0);

            delta = (double)in[i + (y + hsizey) * nx] - colmean[i];
 
            colvarsum[i] += 
            (npixcol - 1.0) / npixcol * delta * delta;
            colmean[i]   += delta / npixcol;
        } else {
            colvarsum[i] = 0.0;
            colmean[i]   = (double)in[i + (y + hsizey) * nx];
        }

        rundelta = colmean[i] - runmean;

        runvarsum += 
            colvarsum[i] + ncol / (ncol + 1.0) * npixcol *
            rundelta * rundelta;
        runmean   += rundelta / (ncol + 1.0);
        
        ncol += 1.0;
            }

            npix += npixy;

            out[x + y * nx] = (OUT_TYPE)sqrt(runvarsum/(npix - 1.0));
        }

        for (; x <= hsizex; x++) {
        double delta;
        double rundelta;

        if (npixcol > 1.0) {
        delta = 
            (double)in[x + hsizex + (y - hsizey - 1) * nx] - 
            colmean[x + hsizex];
 
        colvarsum[x + hsizex] -= 
            npixcol / (npixcol - 1.0) * delta * delta;
        colmean[x + hsizex]   -= delta / (npixcol - 1.0);

        delta = 
            (double)in[x + hsizex + (y + hsizey) * nx] - 
            colmean[x + hsizex];
 
        colvarsum[x + hsizex] += 
            (npixcol - 1.0) / npixcol * delta * delta;
        colmean[x + hsizex]   += delta / npixcol;
        } else {
        colvarsum[x + hsizex] = 0.0;
        colmean[x + hsizex]   = 
            (double)in[x + hsizex + (y + hsizey) * nx];
        }

        rundelta = colmean[x + hsizex] - runmean;

        runvarsum += 
        colvarsum[x + hsizex] + ncol / (ncol + 1.0) * npixcol *
        rundelta * rundelta;
        runmean   += rundelta / (ncol + 1.0);

        ncol += 1.0;
            npix += npixy;

            out[x + y * nx] = (OUT_TYPE)sqrt(runvarsum/(npix - 1.0));
        }

#ifdef ALLOW_RECIPROCAL
        rnpix = (double)1.0/npix;
#endif
     
        for (; x < nx-hsizex; x++) {
        double delta;
        double rundelta;

        if (npixcol > 1.0) {
        delta = 
            (double)in[x + hsizex + (y - hsizey - 1) * nx] - 
            colmean[x + hsizex];
 
        colvarsum[x + hsizex] -= 
            npixcol / (npixcol - 1.0) * delta * delta;
        colmean[x + hsizex]   -= delta / (npixcol - 1.0);

        delta = 
            (double)in[x + hsizex + (y + hsizey) * nx] - 
            colmean[x + hsizex];
 
        colvarsum[x + hsizex] += 
            (npixcol - 1.0) / npixcol * delta * delta;
        colmean[x + hsizex]   += delta / npixcol;
        } else {
        colvarsum[x + hsizex] = 0.0;
        colmean[x + hsizex]   = 
            (double)in[x + hsizex + (y + hsizey) * nx];
        }

        if (ncol > 1.0) {
        rundelta = colmean[x - hsizex - 1] - runmean;

        runvarsum -= 
            colvarsum[x - hsizex - 1] + 
            ncol / (ncol - 1.0) * npixcol * rundelta * rundelta;
        runmean   -= rundelta / (ncol - 1.0);

        rundelta = colmean[x + hsizex] - runmean;

        runvarsum += 
            colvarsum[x + hsizex] + 
            (ncol - 1.0) / ncol * npixcol * rundelta * rundelta;
        runmean   += rundelta / ncol;
        } else {
        runvarsum = colvarsum[x + hsizex];
        }

            out[x + y * nx] = (OUT_TYPE)sqrt(runvarsum/(npix - 1.0));
        }

        for (; x < nx; x++) {
        const double rundelta = colmean[x - hsizex - 1] - runmean;

        runvarsum -= 
        colvarsum[x - hsizex - 1] + 
        ncol / (ncol - 1.0) * npixcol *    rundelta * rundelta;
        runmean   -= rundelta / (ncol - 1.0);

        ncol -= 1.0;
            npix -= npixy;

            out[x + y * nx] = (OUT_TYPE)sqrt(runvarsum/(npix - 1.0));
        }
    }

    /* Loop over all the last rows, that cannot be covered by the kernel */
    for (;y < ny; y++) {

        double runvarsum = (double)0;
        double runmean   = (double)0;

#ifdef ALLOW_RECIPROCAL
        double rnpix;
#endif

        npixy --;
        npix = (hsizex)*npixy;

    ncol = 0.0;

        for (x = 0; x < 1; x++) {

            /* Update the first 1 + hsizex column sums */
            for (i = 0; i <= hsizex; i ++) {
        const double delta = 
            (double)in[i + (y - hsizey - 1) * nx] - colmean[i];
         double rundelta;

        colvarsum[i] -= 
            npixcol / (npixcol - 1.0) * delta * delta;
        colmean[i]   -= delta / (npixcol - 1.0);

        rundelta = colmean[i] - runmean;

        runvarsum += 
            colvarsum[i] + ncol / (ncol + 1.0) * (npixcol - 1.0) *
            rundelta * rundelta;
        runmean   += rundelta / (ncol + 1.0);

        ncol += 1.0;
            }

            npix += npixy;

            out[x + y * nx] = (OUT_TYPE)sqrt(runvarsum/(npix - 1.0));
        }

        for (; x <= hsizex; x++) {
        const double delta =
        (double)in[x + hsizex + (y - hsizey - 1) * nx] - 
        colmean[x + hsizex];
        double rundelta;
 
        colvarsum[x + hsizex] -= 
            npixcol / (npixcol - 1.0) * delta * delta;
        colmean[x + hsizex]   -= delta / (npixcol - 1.0);

        rundelta = colmean[x + hsizex] - runmean;

        runvarsum += 
        colvarsum[x + hsizex] + ncol / (ncol + 1.0) * (npixcol - 1.0) *
        rundelta * rundelta;
        runmean   += rundelta / (ncol + 1.0);

            npix += npixy;
        ncol += 1.0;

            out[x + y * nx] = (OUT_TYPE)sqrt(runvarsum/(npix - 1.0));
        }

#ifdef ALLOW_RECIPROCAL
        rnpix = (double)1.0/npix;
#endif
    
        for (; x < nx-hsizex; x++) {
        const double delta =
        (double)in[x + hsizex + (y - hsizey - 1) * nx] - 
        colmean[x + hsizex];
        double rundelta;
 
        colvarsum[x + hsizex] -= 
        npixcol / (npixcol - 1.0) * delta * delta;
        colmean[x + hsizex]   -= delta / (npixcol - 1.0);

        if (ncol > 1.0) {
        rundelta = colmean[x - hsizex - 1] - runmean;

        runvarsum -= 
            colvarsum[x - hsizex - 1] + 
            ncol / (ncol - 1.0) * (npixcol - 1.0) * rundelta * rundelta;
        runmean   -= rundelta / (ncol - 1.0);

        rundelta = colmean[x + hsizex] - runmean;

        runvarsum += 
            colvarsum[x + hsizex] + 
            (ncol - 1.0) / ncol * (npixcol - 1.0) * rundelta * rundelta;
        runmean   += rundelta / ncol;
        } else {
        runvarsum = colvarsum[x + hsizex];
        }

            out[x + y * nx] = (OUT_TYPE)sqrt(runvarsum/(npix - 1.0));
        }

        for (; x < nx; x++) {
        const double rundelta = colmean[x - hsizex - 1] - runmean;

        runvarsum -= 
        colvarsum[x - hsizex - 1] + 
        ncol / (ncol - 1.0) * (npixcol - 1.0) * rundelta * rundelta;
        runmean   -= rundelta / (ncol - 1.0);

        ncol -= 1.0;
            npix -= npixy;

            out[x + y * nx] = (OUT_TYPE)sqrt(runvarsum/(npix - 1.0));
        }

    npixcol -= 1.0;
    }

    cpl_free(colvarsum);
    cpl_free(colmean);

    return;

}

/*----------------------------------------------------------------------------*/
/**
   @internal
   @ingroup cpl_image
   @brief  Average filter without bad pixels
   @param  vout  The pixel buffer to hold the filtered result
   @param  vin   The input pixel buffer to be filtered
   @param  nx    image width
   @param  ny    image height
   @param  rx    filtering half-size in x
   @param  ry    filtering half-size in y
   @return void
   @note FIXME: Lots of code duplication inside
   
*/
/*----------------------------------------------------------------------------*/
static void
ADDTYPE_TWO(cpl_filter_average)(void * vout, const void * vin,
                                 int nx, int ny, int hsizex, int hsizey,
                                 cpl_border_mode border_mode)
{

    OUT_TYPE * out = (OUT_TYPE*)vout;
    const IN_TYPE  * in  = (const IN_TYPE*)vin;

    ACCU_TYPE * colsum;
    int npix, npixy;
    int x, y, i;

    assert(out != NULL);
    assert(in != NULL);

    assert( hsizex >= 0 );
    assert( hsizey >= 0 );
    assert( 2*hsizex + 1 <= nx ); /* Caller does not (need to) support this */
    assert( 2*hsizey + 1 <= ny ); /* Caller does not (need to) support this */

    /* FIXME: Support more modes ? */
    assert( border_mode == CPL_BORDER_FILTER );

    /* The number of pixels prior to the first average */
    npixy = hsizey;
    npix = (1+hsizex) * npixy;

    /* Initialize the column sums */
    colsum = cpl_calloc(nx, sizeof(*colsum));
    for (y=0; y < hsizey; y++) {
        for (i=0; i < nx; i++) {
            colsum[i] += in[i + y * nx];
        }
    }

    /* Loop over all the first rows, that cannot be covered by the kernel */
    for (y = 0; y <= hsizey; y++) {

        ACCU_TYPE runsum = (ACCU_TYPE)0;

#ifdef ALLOW_RECIPROCAL
        ACCU_TYPE rnpix;
#endif

        npixy ++;
        npix = (hsizex)*npixy;

        for (x = 0; x < 1; x ++) {

            /* Update the first 1 + hsizex column sums */
            for (i = 0; i <= hsizex; i ++) {
              colsum[i] += in[i + (y + hsizey) * nx];
              runsum += colsum[i];
            }

            npix += npixy;

            out[x + y * nx] = (OUT_TYPE)(runsum/npix);
        }

        for (; x <= hsizex; x ++) {

            /* Update the new colsum */
            colsum[x + hsizex] += in[x + hsizex + (y + hsizey) * nx];

            /* Update the running sum */
            runsum += colsum[x + hsizex];

            npix += npixy;

            out[x + y * nx] = (OUT_TYPE)(runsum/npix);
        }

#ifdef ALLOW_RECIPROCAL
        rnpix = (ACCU_TYPE)1.0/npix;
#endif

        for (; x < nx-hsizex; x ++) {

            /* Update the new colsum */
            colsum[x + hsizex] += in[x + hsizex + (y + hsizey) * nx];


            /* Update the running sum */
            runsum -= colsum[x - hsizex - 1];
            runsum += colsum[x + hsizex];

            out[x + y * nx] = (OUT_TYPE)(RUNSUM_MEAN);
        }

        for (; x < nx; x++) {

           /* Update the running sum */
            runsum -= colsum[x - hsizex - 1];

            npix -= npixy;

            out[x + y * nx] = (OUT_TYPE)(runsum/npix);
        }
    }

    /* Loop over the central rows */
    for (;y < ny-hsizey; y++) {

        ACCU_TYPE runsum = (ACCU_TYPE)0;

#ifdef ALLOW_RECIPROCAL
        ACCU_TYPE rnpix;
#endif
        npix = (hsizex)*npixy;

        for (x = 0; x < 1; x++) {

            /* Update the first 1 + hsizex column sums */
            for (i = 0; i <= hsizex; i ++) {
              colsum[i] -= in[i + (y - hsizey - 1) * nx];
              colsum[i] += in[i + (y + hsizey) * nx];
              runsum += colsum[i];
            }

            npix += npixy;

            out[x + y * nx] = (OUT_TYPE)(runsum/npix);
        }

        for (; x <= hsizex; x++) {

            /* Update the new colsum */
            colsum[x + hsizex] -= in[x + hsizex + (y - hsizey - 1) * nx];
            colsum[x + hsizex] += in[x + hsizex + (y + hsizey) * nx];

            /* Update the running sum */
            runsum += colsum[x + hsizex];

            npix += npixy;

            out[x + y * nx] = (OUT_TYPE)(runsum/npix);
        }

#ifdef ALLOW_RECIPROCAL
        rnpix = (ACCU_TYPE)1.0/npix;
#endif
     
        for (; x < nx-hsizex; x++) {

            /* Update the new colsum */
            colsum[x + hsizex] -= in[x + hsizex + (y - hsizey - 1) * nx];
            colsum[x + hsizex] += in[x + hsizex + (y + hsizey) * nx];


            /* Update the running sum */
            runsum -= colsum[x - hsizex - 1];
            runsum += colsum[x + hsizex];

            out[x + y * nx] = (OUT_TYPE)(RUNSUM_MEAN);
        }

        for (; x < nx; x++) {

           /* Update the running sum */
            runsum -= colsum[x - hsizex - 1];

            npix -= npixy;

            out[x + y * nx] = (OUT_TYPE)(runsum/npix);
        }
    }

    /* Loop over all the last rows, that cannot be covered by the kernel */
    for (;y < ny; y++) {

        ACCU_TYPE runsum = (ACCU_TYPE)0;

#ifdef ALLOW_RECIPROCAL
        ACCU_TYPE rnpix;
#endif

        npixy --;
        npix = (hsizex)*npixy;

        for (x = 0; x < 1; x++) {

            /* Update the first 1 + hsizex column sums */
            for (i = 0; i <= hsizex; i ++) {
              colsum[i] -= in[i + (y - hsizey - 1) * nx];
              runsum += colsum[i];
            }

            npix += npixy;

            out[x + y * nx] = (OUT_TYPE)(runsum/npix);
        }

        for (; x <= hsizex; x++) {

            /* Update the new colsum */
            colsum[x + hsizex] -= in[x + hsizex + (y - hsizey - 1) * nx];

            /* Update the running sum */
            runsum += colsum[x + hsizex];

            npix += npixy;

            out[x + y * nx] = (OUT_TYPE)(runsum/npix);
        }

#ifdef ALLOW_RECIPROCAL
        rnpix = (ACCU_TYPE)1.0/npix;
#endif
    
        for (; x < nx-hsizex; x++) {

            /* Update the new colsum */
            colsum[x + hsizex] -= in[x + hsizex + (y - hsizey - 1) * nx];


            /* Update the running sum */
            runsum -= colsum[x - hsizex - 1];
            runsum += colsum[x + hsizex];

            out[x + y * nx] = (OUT_TYPE)(RUNSUM_MEAN);
        }

        for (; x < nx; x++) {

           /* Update the running sum */
            runsum -= colsum[x - hsizex - 1];

            npix -= npixy;

            out[x + y * nx] = (OUT_TYPE)(runsum/npix);
        }
    }

    cpl_free(colsum);

    return;

}

/*----------------------------------------------------------------------------*/
/**
   @internal
   @ingroup cpl_image
   @brief  Average filter with bad pixels
   @param  vout  The pixel buffer to hold the filtered result
   @param  pbpm  *pbpm is the output bpm, or NULL if still to be created
   @param  vin   The input pixel buffer to be filtered
   @param  bpm   The input bad pixel map, or NULL if none
   @param  pmask The window
   @param  nx    image width
   @param  ny    image height
   @param  rx    filtering half-size in x
   @param  ry    filtering half-size in y
   @return void
   @note FIXME: Modified from cpl_filter_average
   
*/
/*----------------------------------------------------------------------------*/
static void
ADDTYPE_TWO(cpl_filter_average_bpm)(void * vout, cpl_binary ** pbpm,
                                    const void * vin, const cpl_binary * bpm,
                                    int nx, int ny, int hsizex, int hsizey,
                                    cpl_border_mode border_mode)
{

    OUT_TYPE * out = (OUT_TYPE*)vout;
    const IN_TYPE  * in  = (const IN_TYPE*)vin;
    ACCU_TYPE * colsum;
    int * colpix;
    int npix;
    int x, y, i;

    assert(out != NULL);
    assert(pbpm != NULL);
    assert(in != NULL);
    assert(bpm != NULL);

    assert( hsizex >= 0 );
    assert( hsizey >= 0 );
    assert( 2*hsizex + 1 <= nx ); /* Caller does not (need to) support this */
    assert( 2*hsizey + 1 <= ny ); /* Caller does not (need to) support this */

    /* FIXME: Support more modes ? */
    assert( border_mode == CPL_BORDER_FILTER );

    /* Initialize the column sums */
    colsum = cpl_calloc(nx, sizeof(*colsum));
    /* Initialize the counter of column elements (variable with bad pixels) */
    colpix = cpl_calloc(nx, sizeof(*colsum));
    for (y=0; y < hsizey; y++) {
        for (i=0; i < nx; i++) {
            if (!bpm[i + y * nx]) {
                colsum[i] += in[i + y * nx];
                colpix[i]++;
            }
        }
    }

    /* Loop over all the first rows, that cannot be covered by the kernel */
    for (y = 0; y <= hsizey; y++) {

        ACCU_TYPE runsum = (ACCU_TYPE)0;

#ifdef ALLOW_RECIPROCAL
        ACCU_TYPE rnpix = 0.0; /* Prevent (false) uninit warning */
#endif

        npix = 0;

        for (x = 0; x < 1; x ++) {

            /* Update the first 1 + hsizex column sums */
            for (i = 0; i <= hsizex; i ++) {
                if (!bpm[i + (y + hsizey) * nx]) {
                    colsum[i] += in[i + (y + hsizey) * nx];
                    runsum += colsum[i];
                    colpix[i]++;
                    npix += colpix[i];
                }
            }

            if (npix) {
                out[x + y * nx] = (OUT_TYPE)(runsum/npix);
            } else {
                /* Flag as bad - and set to zero */
                if (*pbpm == NULL) *pbpm
                    = (cpl_binary*)cpl_calloc(nx * ny, sizeof(cpl_binary));
                (*pbpm)[x + y * nx] = CPL_BINARY_1;
                out[x + y * nx] = (OUT_TYPE)0;
            }
        }

        for (; x <= hsizex; x ++) {

            if (!bpm[x + hsizex + (y + hsizey) * nx]) {
                /* Update the new colsum */
                colsum[x + hsizex] += in[x + hsizex + (y + hsizey) * nx];

                /* Update the running sum */
                runsum += colsum[x + hsizex];

                colpix[x + hsizex]++;
                npix += colpix[x + hsizex];

            }
            if (npix) {
                out[x + y * nx] = (OUT_TYPE)(runsum/npix);
            } else {
                /* Flag as bad - and set to zero */
                if (*pbpm == NULL) *pbpm
                    = (cpl_binary*)cpl_calloc(nx * ny, sizeof(cpl_binary));
                (*pbpm)[x + y * nx] = CPL_BINARY_1;
                out[x + y * nx] = (OUT_TYPE)0;
            }
        }

#ifdef ALLOW_RECIPROCAL
        rnpix = (ACCU_TYPE)1.0/npix;
#endif

        for (; x < nx-hsizex; x ++) {

            if (!bpm[x + hsizex + (y + hsizey) * nx]) {
                /* Update the new colsum */
                colsum[x + hsizex] += in[x + hsizex + (y + hsizey) * nx];
                colpix[x + hsizex]++;
            }
            if (colsum[x + hsizex] != colpix[x - hsizex - 1]) {
                npix += colpix[x + hsizex] - colpix[x - hsizex - 1];
#ifdef ALLOW_RECIPROCAL
                if (npix) rnpix = (ACCU_TYPE)1.0/npix;
#endif
            }

            /* Update the running sum */
            runsum -= colsum[x - hsizex - 1];
            runsum += colsum[x + hsizex];

            if (npix) {
                out[x + y * nx] = (OUT_TYPE)(RUNSUM_MEAN);
            } else {
                /* Flag as bad - and set to zero */
                if (*pbpm == NULL) *pbpm
                    = (cpl_binary*)cpl_calloc(nx * ny, sizeof(cpl_binary));
                (*pbpm)[x + y * nx] = CPL_BINARY_1;
                out[x + y * nx] = (OUT_TYPE)0;
            }
        }

        for (; x < nx; x++) {

           /* Update the running sum */
            runsum -= colsum[x - hsizex - 1];

            npix -= colpix[x - hsizex - 1];

            if (npix) {
                out[x + y * nx] = (OUT_TYPE)(runsum/npix);
            } else {
                /* Flag as bad - and set to zero */
                if (*pbpm == NULL) *pbpm
                    = (cpl_binary*)cpl_calloc(nx * ny, sizeof(cpl_binary));
                (*pbpm)[x + y * nx] = CPL_BINARY_1;
                out[x + y * nx] = (OUT_TYPE)0;
            }
        }
    }

    /* Loop over the central rows */
    for (;y < ny-hsizey; y++) {

        ACCU_TYPE runsum = (ACCU_TYPE)0;

#ifdef ALLOW_RECIPROCAL
        ACCU_TYPE rnpix = (ACCU_TYPE)0; /* Prevent (false) uninit warning */
#endif
        npix = 0;

        for (x = 0; x < 1; x++) {

            /* Update the first 1 + hsizex column sums */
            for (i = 0; i <= hsizex; i ++) {
                if (!bpm[i + (y - hsizey - 1) * nx]) {
                    colsum[i] -= in[i + (y - hsizey - 1) * nx];
                    colpix[i]--;
                }
                if (!bpm[i + (y + hsizey) * nx]) {
                    colsum[i] += in[i + (y + hsizey) * nx];
                    colpix[i]++;
                }
                runsum += colsum[i];
                npix += colpix[i];
            }

            if (npix) {
                out[x + y * nx] = (OUT_TYPE)(runsum/npix);
            } else {
                /* Flag as bad - and set to zero */
                if (*pbpm == NULL) *pbpm
                    = (cpl_binary*)cpl_calloc(nx * ny, sizeof(cpl_binary));
                (*pbpm)[x + y * nx] = CPL_BINARY_1;
                out[x + y * nx] = (OUT_TYPE)0;
            }
        }

        for (; x <= hsizex; x++) {

            /* Update the new colsum */
            if (!bpm[x + hsizex + (y - hsizey - 1) * nx]) {
                colsum[x + hsizex] -= in[x + hsizex + (y - hsizey - 1) * nx];
                colpix[x + hsizex]--;
            }
            if (!bpm[x + hsizex + (y + hsizey) * nx]) {
                colsum[x + hsizex] += in[x + hsizex + (y + hsizey) * nx];
                colpix[x + hsizex]++;
            }

            /* Update the running sum */
            runsum += colsum[x + hsizex];
            npix += colpix[x + hsizex];

            if (npix) {
                out[x + y * nx] = (OUT_TYPE)(runsum/npix);
            } else {
                /* Flag as bad - and set to zero */
                if (*pbpm == NULL) *pbpm
                    = (cpl_binary*)cpl_calloc(nx * ny, sizeof(cpl_binary));
                (*pbpm)[x + y * nx] = CPL_BINARY_1;
                out[x + y * nx] = (OUT_TYPE)0;
            }
        }

#ifdef ALLOW_RECIPROCAL
        if (npix > (ACCU_TYPE)0.0)
            rnpix = (ACCU_TYPE)1.0/npix;
#endif
     
        for (; x < nx-hsizex; x++) {

            /* Update the new colsum */
            if (!bpm[x + hsizex + (y - hsizey - 1) * nx]) {
                colsum[x + hsizex] -= in[x + hsizex + (y - hsizey - 1) * nx];
                colpix[x + hsizex]--;
            }
            if (!bpm[x + hsizex + (y + hsizey) * nx]) {
                colsum[x + hsizex] += in[x + hsizex + (y + hsizey) * nx];
                colpix[x + hsizex]++;
            }

            /* Update the running sum */
            runsum -= colsum[x - hsizex - 1];
            runsum += colsum[x + hsizex];

            if (colsum[x + hsizex] != colpix[x - hsizex - 1]) {
                npix += colpix[x + hsizex] - colpix[x - hsizex - 1];
#ifdef ALLOW_RECIPROCAL
                if (npix) rnpix = (ACCU_TYPE)1.0/npix;
#endif
            }

            if (npix) {
                out[x + y * nx] = (OUT_TYPE)(RUNSUM_MEAN);
            } else {
                /* Flag as bad - and set to zero */
                if (*pbpm == NULL) *pbpm
                    = (cpl_binary*)cpl_calloc(nx * ny, sizeof(cpl_binary));
                (*pbpm)[x + y * nx] = CPL_BINARY_1;
                out[x + y * nx] = (OUT_TYPE)0;
            }
        }

        for (; x < nx; x++) {

           /* Update the running sum */
            runsum -= colsum[x - hsizex - 1];
            npix   -= colpix[x - hsizex - 1];

            if (npix) {
                out[x + y * nx] = (OUT_TYPE)(runsum/npix);
            } else {
                /* Flag as bad - and set to zero */
                if (*pbpm == NULL) *pbpm
                    = (cpl_binary*)cpl_calloc(nx * ny, sizeof(cpl_binary));
                (*pbpm)[x + y * nx] = CPL_BINARY_1;
                out[x + y * nx] = (OUT_TYPE)0;
            }
        }
    }

    /* Loop over all the last rows, that cannot be covered by the kernel */
    for (;y < ny; y++) {

        ACCU_TYPE runsum = (ACCU_TYPE)0;

#ifdef ALLOW_RECIPROCAL
        ACCU_TYPE rnpix = (ACCU_TYPE)0; /* Prevent (false) uninit warning */
#endif

        npix = (ACCU_TYPE)0;

        for (x = 0; x < 1; x++) {

            /* Update the first 1 + hsizex column sums */
            for (i = 0; i <= hsizex; i ++) {
                if (!bpm[i + (y - hsizey - 1) * nx]) {
                    colsum[i] -= in[i + (y - hsizey - 1) * nx];
                    colpix[i]--;
                }
              runsum += colsum[i];
              npix += colpix[i];
            }

            if (npix) {
                out[x + y * nx] = (OUT_TYPE)(runsum/npix);
            } else {
                /* Flag as bad - and set to zero */
                if (*pbpm == NULL) *pbpm
                    = (cpl_binary*)cpl_calloc(nx * ny, sizeof(cpl_binary));
                (*pbpm)[x + y * nx] = CPL_BINARY_1;
                out[x + y * nx] = (OUT_TYPE)0;
            }
        }

        for (; x <= hsizex; x++) {

            /* Update the new colsum */
                if (!bpm[x + hsizex + (y - hsizey - 1) * nx]) {
                    colsum[x + hsizex] -= in[x + hsizex + (y - hsizey - 1) * nx];
                    colpix[x + hsizex]--;
                }

            /* Update the running sum */
            runsum += colsum[x + hsizex];
            npix   += colpix[x + hsizex];

            if (npix) {
                out[x + y * nx] = (OUT_TYPE)(runsum/npix);
            } else {
                /* Flag as bad - and set to zero */
                if (*pbpm == NULL) *pbpm
                    = (cpl_binary*)cpl_calloc(nx * ny, sizeof(cpl_binary));
                (*pbpm)[x + y * nx] = CPL_BINARY_1;
                out[x + y * nx] = (OUT_TYPE)0;
            }
        }

#ifdef ALLOW_RECIPROCAL
        if (npix) rnpix = (ACCU_TYPE)1.0/npix;
#endif
    
        for (; x < nx-hsizex; x++) {

            /* Update the new colsum */
            if (!bpm[x + hsizex + (y - hsizey - 1) * nx]) {
                colsum[x + hsizex] -= in[x + hsizex + (y - hsizey - 1) * nx];
                colpix[x + hsizex]--;
            }


            /* Update the running sum */
            runsum -= colsum[x - hsizex - 1];
            runsum += colsum[x + hsizex];

            if (colsum[x + hsizex] != colpix[x - hsizex - 1]) {
                npix += colpix[x + hsizex] - colpix[x - hsizex - 1];
#ifdef ALLOW_RECIPROCAL
                if (npix) rnpix = (ACCU_TYPE)1.0/npix;
#endif
            }

            if (npix) {
                out[x + y * nx] = (OUT_TYPE)(RUNSUM_MEAN);
            } else {
                /* Flag as bad - and set to zero */
                if (*pbpm == NULL) *pbpm
                    = (cpl_binary*)cpl_calloc(nx * ny, sizeof(cpl_binary));
                (*pbpm)[x + y * nx] = CPL_BINARY_1;
                out[x + y * nx] = (OUT_TYPE)0;
            }
        }

        for (; x < nx; x++) {

           /* Update the running sum */
            runsum -= colsum[x - hsizex - 1];
            npix   -= colpix[x - hsizex - 1];

            if (npix) {
                out[x + y * nx] = (OUT_TYPE)(runsum/npix);
            } else {
                /* Flag as bad - and set to zero */
                if (*pbpm == NULL) *pbpm
                    = (cpl_binary*)cpl_calloc(nx * ny, sizeof(cpl_binary));
                (*pbpm)[x + y * nx] = CPL_BINARY_1;
                out[x + y * nx] = (OUT_TYPE)0;
            }
        }
    }

    cpl_free(colsum);
    cpl_free(colpix);

    return;

}


/*----------------------------------------------------------------------------*/
/**
   @internal
   @ingroup cpl_image
   @brief  Linear filter with bad pixels
   @param  vout    The pixel buffer to hold the filtered result
   @param  ppbpm   *ppbpm is the bpm, or NULL if still to be created
   @param  vin     The input pixel buffer to be filtered
   @param  bpm     The input bad pixel map, or NULL if none
   @param  pkernel The kernel
   @param  nx      image width
   @param  ny      image height
   @param  rx      filtering half-size in x
   @param  ry      filtering half-size in y
   @return void
   @see cpl_filter_average_slow
   @note FIXME: Modified from cpl_filter_average_slow
   
*/
/*----------------------------------------------------------------------------*/
static void
ADDTYPE_TWO(cpl_filter_linear_slow)(void * vout, cpl_binary ** ppbpm,
                                    const void * vin,
                                    const cpl_binary * bpm,
                                    const double * pkernel,
                                    int nx, int ny,
                                    int rx, int ry,
                                    cpl_border_mode mode)
{

    if (mode == CPL_BORDER_FILTER) {
        OUT_TYPE * out = (OUT_TYPE*)vout;
        const IN_TYPE  * in  = (const IN_TYPE*)vin;
        int        x, y;

        for (y = 0; y < ny; y++) {
            const int y_1 = CX_MAX(0,    y - ry);
            const int y2  = CX_MIN(ny-1, y + ry);
            for (x = 0; x < nx; x++) {
                const int          x1 = CX_MAX(0,    x - rx);
                const int          x2 = CX_MIN(nx-1, x + rx);
                const IN_TYPE    * inj  = in + y_1 * nx;
                const cpl_binary * bpmj = bpm ? bpm + y_1 * nx : NULL;
                /* Point to last (bottom) kernel row for first (bottom)
                   image row */
                /* As a difference from cpl_filter_average_slow,
                   element (i,j) in a mask buffer corresponds to
                   element (i, n-1-j) in a matrix buffer. Thus a
                   matrix buffer has to be accessed in reverse row
                   order */
                const double     * pkernelj = pkernel
                    + (rx - x) + (ry + y - y_1) * (2 * rx + 1);
                double             sum    = 0.0;
                double             weight = 0.0;
                int i, j;

                for (j = y_1; j <= y2; j++, inj += nx, bpmj += nx,
                         pkernelj -= 2 * rx + 1) {
                    for (i = x1; i <= x2; i++) {
                        if (bpm == NULL || !bpmj[i]) {
                            sum += pkernelj[i] * (double)inj[i];
                            weight += fabs(pkernelj[i]);
                        }
                    }
                }
                if (weight > 0.0) {
                    sum /= weight;
                } else {
                    /* Flag as bad - and set to zero */
                    if (*ppbpm == NULL) {
                        *ppbpm = cpl_calloc(nx * ny, sizeof(cpl_binary));
                    }
                    (*ppbpm)[x + y * nx] = CPL_BINARY_1;
                }
                out[x + y * nx] = (OUT_TYPE)sum;
            }
        }
    } else {
        /* FIXME: Support more modes ? */
        (void)cpl_error_set(cpl_func, CPL_ERROR_UNSUPPORTED_MODE);
    }

    return;
}


/*----------------------------------------------------------------------------*/
/**
   @internal
   @ingroup cpl_image
   @brief  Morpho filter with bad pixels
   @param  vout    The pixel buffer to hold the filtered result
   @param  ppbpm   *ppbpm is the bpm, or NULL if still to be created
   @param  vin     The input pixel buffer to be filtered
   @param  bpm     The input bad pixel map, or NULL if none
   @param  pkernel The kernel
   @param  nx      image width
   @param  ny      image height
   @param  rx      filtering half-size in x
   @param  ry      filtering half-size in y
   @return void
   @see cpl_filter_average_slow
   @note FIXME: Modified from cpl_filter_linear_slow + cpl_filter_median_slow
   
*/
/*----------------------------------------------------------------------------*/
static void
ADDTYPE_TWO(cpl_filter_morpho_slow)(void * vout, cpl_binary ** ppbpm,
                                    const void * vin,
                                    const cpl_binary * bpm,
                                    const double * pkernel,
                                    int nx, int ny,
                                    int rx, int ry,
                                    cpl_border_mode mode)
{

    if (mode == CPL_BORDER_FILTER) {
        OUT_TYPE * out = (OUT_TYPE*)vout;
        const IN_TYPE  * in  = (const IN_TYPE*)vin;
        IN_TYPE window[(2 * rx + 1) * (2 * ry + 1)];
        int        x, y;

        for (y = 0; y < ny; y++) {
            const int y_1 = CX_MAX(0,    y - ry);
            const int y2  = CX_MIN(ny-1, y + ry);
            for (x = 0; x < nx; x++) {
                const int          x1 = CX_MAX(0,    x - rx);
                const int          x2 = CX_MIN(nx-1, x + rx);
                const IN_TYPE    * inj  = in + y_1 * nx;
                const cpl_binary * bpmj = bpm ? bpm + y_1 * nx : NULL;
                double             sum    = 0.0;
                double             weight = 0.0;
                int i, j;
                int k = 0;

                /* FIXME: In the sorted array, only one row has to be updated
                   for each new pixel, see the fast median algorithm */

                for (j = y_1; j <= y2; j++, inj += nx, bpmj += nx) {
                    for (i = x1; i <= x2; i++) {
                        if (bpm == NULL || !bpmj[i]) {
                            window[k] = inj[i];
                            weight += fabs(pkernel[k]);
                            k++;
                        }
                    }
                }
                if (weight > 0.0) {
                    /* Sort array with standard sorting routine */
                    /* assert(k > 0); */
                    ADDTYPE(cpl_tools_sort)(window, k);
                    for (; k--;) {
                        sum += pkernel[k] * (double)window[k];
                    }
                    sum /= weight;
                } else {
                    /* Flag as bad - and set to zero */
                    if (*ppbpm == NULL) {
                        *ppbpm = cpl_calloc(nx * ny, sizeof(cpl_binary));
                    }
                    (*ppbpm)[x + y * nx] = CPL_BINARY_1;
                }
                out[x + y * nx] = (OUT_TYPE)sum;
            }
        }
    } else {
        /* FIXME: Support more modes ? */
        (void)cpl_error_set(cpl_func, CPL_ERROR_UNSUPPORTED_MODE);
    }

    return;
}

/* @endcond
 */

#undef ACCU_TYPE
#undef ACCU_FLOATTYPE
#undef IN_TYPE
#undef OUT_TYPE
#undef ACCU_TYPE_IS_INT
#undef ALLOW_RECIPROCAL
#undef RUNSUM_MEAN

